from __future__ import annotations

from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional, Union, List, Tuple, Dict

from obspy import UTCDateTime
from obspy.core import Stream

from PySide2 import QtCore
from PySide2 import QtWidgets

from sohstationviewer.controller.util import display_tracking_info
from sohstationviewer.conf import constants
from sohstationviewer.view.plotting.gps_plot.gps_point import GPSPoint
from sohstationviewer.view.util.enums import LogType
from sohstationviewer.database.process_db import execute_db
from sohstationviewer.model.handling_data import (
    combine_traces_except_overlap, sort_data, retrieve_gaps_from_stream_header)


class ProcessingDataError(Exception):
    def __init__(self, msg):
        self.message = msg


class ThreadStopped(Exception):
    """
    An exception that is raised when the user requests for the data loader
    thread to be stopped.
    """
    def __init__(self, *args, **kwargs):
        self.args = (args, kwargs)


class DataTypeModel():
    def __init__(self, data_type, tracking_box: QtWidgets.QTextBrowser,
                 folder: str, list_of_rt130_paths: List[Path] = [],
                 req_wf_chans: Union[List[str], List[int]] = [],
                 req_soh_chans: List[str] = [],
                 read_start: Optional[float] = UTCDateTime(0).timestamp,
                 read_end: Optional[float] = UTCDateTime().timestamp,
                 include_mp123zne: bool = False,
                 include_mp456uvw: bool = False,
                 creator_thread: Optional[QtCore.QThread] = None,
                 notification_signal: Optional[QtCore.Signal] = None,
                 pause_signal: Optional[QtCore.Signal] = None,
                 *args, **kwargs):
        """
        Super class for different data type to process data from data files

        :param data_type: type of the object
        :param tracking_box: widget to display tracking info
        :param folder: path to the folder of data
        :param list_of_rt130_paths: path to the folders of RT130 data
        :param req_wf_chans: requested waveform channel list
        :param req_soh_chans: requested SOH channel list
        :param read_start: requested start time to read
        :param read_end: requested end time to read
        :param include_mp123zne: if mass position channels 1,2,3 are requested
        :param include_mp456uvw: if mass position channels 4,5,6 are requested
        :param creator_thread: the thread the current DataTypeModel instance is
            being created in. If None, the DataTypeModel instance is being
            created in the main thread
        :param notification_signal: signal used to send notifications to the
            main thread. Only not None when creator_thread is not None
        :param pause_signal: signal used to notify the main thread that the
            data loader is paused.
        """
        self.data_type = data_type
        self.tracking_box = tracking_box
        self.dir = folder
        self.list_of_rt130_paths = list_of_rt130_paths
        self.req_soh_chans = req_soh_chans
        self.req_wf_chans = req_wf_chans
        self.read_start = read_start
        self.read_end = read_end
        self.include_mp123zne = include_mp123zne
        self.include_mp456uvw = include_mp456uvw
        if creator_thread is None:
            err_msg = (
                'A signal is not None while running in main thread'
            )
            assert notification_signal is None, err_msg
            assert pause_signal is None, err_msg
            self.creator_thread = QtCore.QThread()
        else:
            self.creator_thread = creator_thread
        self.notification_signal = notification_signal
        self.pause_signal = pause_signal

        """
        stream_header_by_key_chan: stream header by key, chan_id to get key
        list, gaps by sta_id, nets by sta_id, channels by sta_id
        """
        self.stream_header_by_key_chan: Dict[str, Dict[str, Stream]] = {}
        """
        processing_log: [(message, type)] - record the progress of processing
        """
        self.processing_log: List[Tuple[str, LogType]] = []

        DataKey = Union[Tuple[str, str], str]

        """
        Log data: info from log channels, soh messages, text file in dict:
        {'TEXT': [str,], key:{chan_id: [str,],},}
        In which 'TEXT': is the chan_id given by sohview for text only file.
        Note: log_data for RT130's dataset has only one channel: SOH
        """
        self.log_data = {'TEXT': []}  # noqa

        """
        waveform_data: data of waveform in dict:
        {set_key: {
            chan_id - str: {
                'samplerate': Sample rate of the data - float
                'tracesInfo': [{
                    'samplerate': Sample rate of the data - float
                    'startTmEpoch': Start epoch time of the trace - float
                    'endTmEpoch': End epoch time of the trace - float
                    'size': Size of data - int
                    'time': time of trace for channel with spr<=1
                        - np.array of float
                    'data': data of trace for channel with spr<=1
                        - np.array of float
                    'data_f': np.memmap's file path that save data for channel
                        with spr>1 - str
                    },]
                'times': times that has been trimmed and down- sampled for
                    plotting - np.array of float,
                'data': data that has been trimmed and down-sampled for
                    plotting - np.array of float/int,
                'chan_db_info': the plotting parameters got from database
                    for this channel - dict,
                ax_wf: axes to draw the channel in WaveformWidget, to
                    combine 2 channels in need - matplotlib.axes.Axes
            }
        }
        (np.memmap is used instead of data that have spr>1 because they
        are too large and make system crashed when trying to process)
        """
        self.waveform_data = {}

        """
        soh_data: data for SOH info in dict:
        {set_key - str or (str, str): {
            chan_id - str: {
                'reftek': flag to show that the data is reftek to include
                    handling key 'logIdx' in trace - bool,
                'samplerate': sample rate of the data - float,
                'tracesInfo': [{
                    'samplerate': Sample rate of the data - float
                    'startTmEpoch': Start epoch time of the trace - float
                    'endTmEpoch': End epoch time of the trace - float
                    'size': Size of data - int
                    'time': time of trace for channel with spr<=1
                        - np.array of float
                    'data': data of trace for channel with spr<=1
                        - np.array of float
                    'data_f': np.memmap's file path that save data for channel
                        with spr>1 - str
                    'logIdx: soh message line indexes - np.array of int,
                    },]
                times: times that has been trimmed and down- sampled for
                    plotting - np.array of float,
                data: data that has been trimmed and down-sampled for plotting
                    - np.array of float/int,
                logIdx: soh message line indexes that has been trimmed and
                    down-sampled for plotting - np.array of int,
               'chan_db_info': the plotting parameters got from database
                    for this channel - dict,
                ax: axes to draw the channel in PlottingWidget
                    - matplotlib.axes.Axes
                }
            }
        }
        """
        self.soh_data = {}
        """
        mass_pos_data: data for mass position info in the format similar to
        soh_data but there are 2 axes for drawing a mass possition channel,
        one is 'ax' to display in PlottingWidget, the other is 'ax_wf' to
        display in WaveformWidget.
        Channel name for mseed mass position: VP*
        Channel name for reftek mass position: MassPos* (given by SOHView)
                {set_key - str or (str, str): {
            chan_id - str: {
                'tracesInfo': [{
                    'samplerate': Sample rate of the data - float
                    'startTmEpoch': Start epoch time of the trace - float
                    'endTmEpoch': End epoch time of the trace - float
                    'size': Size of data - int
                    'time': time of trace for channel with spr<=1
                        - np.array of float
                    'data': data of trace for channel with spr<=1
                        - np.array of float
                    'data_f': np.memmap's file path that save data for channel
                        with spr>1 - str
                    },]
                times: times that has been trimmed and down- sampled for
                    plotting - np.array of float,
                data: data that has been trimmed and down-sampled for plotting
                    - np.array of float/int,
               'chan_db_info': the plotting parameters got from database
                    for this channel - dict,
                ax: axes to draw the channel in PlottingWidget
                    - matplotlib.axes.Axes,
                ax_wf: axes to draw the channel in WaveformWidget
                    - matplotlib.axes.Axes
                }
            }
        }
        NOTE: Each mass position channel is plotted twice.
        All axes in soh dialog will use key 'ax'.
        All axes in waveform dialog will use key 'ax_wf'.
        A mass position channel will have 2 keys for axes.
        """
        self.mass_pos_data = {}

        """
        data_time: time range of data sets:
        {setKey: [earliestepoch, latestepoch]} - {str: [float, float],}
        """
        self.data_time: Dict[DataKey, List[float]] = {}

        """
        The given data may include more than one data set which is station_id
        in mseed or (unit_id, exp_no) in reftek. User are allow to choose which
        data set to be displayed
        selected_key: str - key of the data set to be displayed
        """
        self.selected_key: Optional[str] = None

        """
        gaps: gaps info in dict:
        {set_key: [list of gaps]} - {str: [[float, float],],}
        """
        self.gaps: Dict[DataKey, List[List[float]]] = {}

        """
         tmp_dir: str - dir to keep memmap files.
         Will be deleted when object is deleted
        """
        self.tmp_dir_obj: TemporaryDirectory = TemporaryDirectory()
        self.tmp_dir = self.tmp_dir_obj.name
        self.save_temp_data_folder_to_database()

        self._pauser = QtCore.QSemaphore()
        self.pause_response = None

        self.gps_points: List[GPSPoint] = []

    def read_folder(self, folder: str) -> Tuple[Dict]:
        """
        Read data from given folder
        :param folder: path to folder to read data
        :return: Tuple of different data dicts
        """
        pass

    def select_key(self) -> Union[str, Tuple[str, str]]:
        """
        Get the key for the data set to process.
        :return: key of the selected data set
        """
        pass

    def processing_data(self):
        if self.creator_thread.isInterruptionRequested():
            raise ThreadStopped()
        self.read_folder(self.dir)

        if self.creator_thread.isInterruptionRequested():
            raise ThreadStopped()
        self.selected_key = self.select_key()
        if self.selected_key is None:
            raise ThreadStopped()

        if self.creator_thread.isInterruptionRequested():
            raise ThreadStopped()
        self.finalize_data()

    def finalize_data(self):
        """
        This function should be called after all folders finish reading to
            + get gaps from stream_header_by_key_chan
            + sort all data
            + combine traces in data and split at overlaps
            + check not found soh channels to give user a warning if needed
            + change data time with default value that are invalid for plotting
                to read_start, read_end.
        """
        self.track_info("Retrieve gaps.", LogType.INFO)
        retrieve_gaps_from_stream_header(
            self.stream_header_by_key_chan,
            self.gaps, self.read_start, self.read_end)

        self.track_info("Sort data.", LogType.INFO)
        self.sort_all_data()
        self.track_info("Combine data.", LogType.INFO)
        self.combine_traces_in_all_data()
        self.check_not_found_soh_channels()
        for key in self.data_time:
            if self.data_time[key] == [constants.HIGHEST_INT, 0]:
                # this happens when there is text or ascii only in the data
                self.data_time[key] = [self.read_start, self.read_end]

    def __del__(self):
        print("delete dataType Object")
        try:
            del self.tmp_dir_obj
        except OSError as e:
            self.track_info(
                "Error deleting %s : %s" % (self.tmp_dir, e.strerror),
                LogType.ERROR)
            print("Error deleting %s : %s" % (self.tmp_dir, e.strerror))
        print("finish deleting")

    def has_data(self) -> bool:
        """
        :return bool - True if there is any data can be read.
                       False if there is no valid data
        """
        if (len(self.log_data) == 1 and len(self.soh_data) == 0 and
                len(self.mass_pos_data) == 0 and len(self.waveform_data) == 0):
            return False
        return True

    def track_info(self, text: str, type: LogType) -> None:
        """
        Display tracking info in tracking_box.
        Add all errors/warnings to processing_log.
        :param text: str - message to display
        :param type: str - type of message (error/warning/info)
        """
        # display_tracking_info updates a QtWidget, which can only be done in
        # the main thread. So, if we are running in a background thread
        # (i.e. self.creator_thread is not None), we need to use signal slot
        # mechanism to ensure that display_tracking_info is run in the main
        # thread.
        if self.notification_signal is None:
            display_tracking_info(self.tracking_box, text, type)
        else:
            self.notification_signal.emit(self.tracking_box, text, type)
        if type != LogType.INFO:
            self.processing_log.append((text, type))

    @classmethod
    def create_data_object(cls, data_type, tracking_box, folder,
                           list_of_rt130_paths,
                           req_wf_chans=[], req_soh_chans=[],
                           read_start=0, read_end=constants.HIGHEST_INT,
                           include_mp123=False, include_mp456=False):
        """
        Create a DataTypeModel object, with the concrete class being based on
        data_type. Run on the same thread as its caller, and so will block the
        GUI if called on the main thread. Do not call this method directly.
        Instead, call the wrapper controller.processing.load_data.

        :param data_type: str - type of data read
        :param tracking_box: QTextBrowser - widget to display tracking info
        :param folder: [str,] - the data directory
        :param list_of_rt130_paths: [Path,]: list of RT130 directories
        :param req_wf_chans: [str,] - requested waveform channel list
        :param req_soh_chans: [str,] - requested soh channel list
        :param read_start: [float,] - start time of read data
        :param read_end: [float,] - finish time of read data
        :param include_mp123: bool - if mass pos 1,2,3 are requested
        :param include_mp456: bool - if mass pos 4,5,6 are requested
        :return: DataTypeModel - object that keep the data read from
            folder
        """
        if data_type == 'RT130':
            from sohstationviewer.model.reftek.reftek import RT130
            data_object = RT130(
                data_type, tracking_box, folder, list_of_rt130_paths,
                reqWFChans=req_wf_chans, reqSOHChans=req_soh_chans,
                readStart=read_start, readEnd=read_end,
                include_mp123=include_mp123, include_mp456=include_mp456)
        else:
            from sohstationviewer.model.mseed.mseed import MSeed
            data_object = MSeed(
                data_type, tracking_box, folder, list_of_rt130_paths,
                reqWFChans=req_wf_chans, reqSOHChans=req_soh_chans,
                readStart=read_start, readEnd=read_end,
                include_mp123=include_mp123, include_mp456=include_mp456)
        return data_object

    def pause(self) -> None:
        """
        Pause the thread this DataTypeModel instance is in. Works by trying
        to acquire a semaphore that is not available, which causes the thread
        to block.

        Note: due to how this is implemented, each call to pause will require
        a corresponding call to unpause. Thus, it is inadvisable to call this
        method more than once.

        Caution: not safe to call in the main thread. Unless a background
        thread releases the semaphore, the whole program will freeze.
        """
        self._pauser.acquire()

    @QtCore.Slot()
    def unpause(self):
        """
        Unpause the thread this DataTypeModel instance is in. Works by trying
        to acquire a semaphore that is not available, which causes the thread
        to block.

        Caution: due to how this is implemented, if unpause is called before
        pause, the thread will not be paused until another call to pause is
        made. Also, like pause, each call to unpause must be matched by another
        call to pause for everything to work.
        """
        self._pauser.release()

    @QtCore.Slot()
    def receive_pause_response(self, response: object):
        """
        Receive a response to a request made to another thread and unpause the
        calling thread.

        :param response: the response to the request made
        :type response: object
        """
        self.pause_response = response
        self.unpause()

    @classmethod
    def get_empty_instance(cls) -> DataTypeModel:
        """
        Create an empty data object. Useful if a DataTypeModel instance is
        needed, but it is undesirable to load a data set. Basically wraps
        __new__().

        :return: an empty data object
        :rtype: DataTypeModel
        """
        return cls.__new__(cls)

    def save_temp_data_folder_to_database(self):
        execute_db(f'UPDATE PersistentData SET FieldValue="{self.tmp_dir}" '
                   f'WHERE FieldName="tempDataDirectory"')

    def check_not_found_soh_channels(self):
        all_chans_meet_req = (
                list(self.soh_data[self.selected_key].keys()) +
                list(self.mass_pos_data[self.selected_key].keys()) +
                list(self.log_data[self.selected_key].keys()))

        not_found_chans = [c for c in self.req_soh_chans
                           if c not in all_chans_meet_req]
        if not_found_chans != []:
            msg = (f"No data found for the following channels: "
                   f"{', '.join( not_found_chans)}")
            self.processing_log.append((msg, LogType.WARNING))

    def combine_times_data_of_traces_w_spr_less_or_equal_1(
            self, sta_data: Dict[str, Dict], data_name: str):
        """
        Create plotting times and data for channels with samplerate
            less than or equal to 1, in which plotting times is
            all traces' time combined together but split where there is
            overlap to prevent plotting line go back and fort,
            and same for plotting data.
        :param sta_data: chan_data of a station by chan_id
        :param data_name: name of data (Waveform, SOH, Mass Position)

        :return: the result plotting times and data of each channel will be
            used to create times and data items of the channel.
        """
        if self.creator_thread.isInterruptionRequested():
            raise ThreadStopped()
        self.track_info(
            f'{data_name}: Combine traces with samplerate < 1', LogType.INFO)

        for chan_id in sta_data:
            chan_data = sta_data[chan_id]
            if chan_data['samplerate'] > 1:
                continue
            new_traces = combine_traces_except_overlap(chan_data['tracesInfo'])
            chan_data['tracesInfo'] = new_traces

    def sort_all_data(self):
        """
        Sort by startTmEpoch on all data: waveform_data, mass_pos_data,
            soh_data.
        Reftek's soh_data won't be sorted here. It has been sorted by time
            because it is created from log data which is sorted in
            prepare_soh_data_from_log_data()
        """
        sort_data(self.waveform_data[self.selected_key])
        sort_data(self.mass_pos_data[self.selected_key])
        try:
            sort_data(self.soh_data[self.selected_key])
        except KeyError:
            # Reftek's SOH trace doesn't have startTmEpoch and
            # actually soh_data consists of only one trace
            pass

    def combine_traces_in_all_data(self):
        """
        Combine times and data of traces that have sample rate less than or
            equal to 1 on waveform_data, mass_position_data and soh_data.
        Reftek's soh_data won't be combined because of its format.
        """
        if self.selected_key in self.waveform_data.keys():
            self.combine_times_data_of_traces_w_spr_less_or_equal_1(
                self.waveform_data[self.selected_key], 'Waveform')
        if self.selected_key in self.mass_pos_data.keys():
            self.combine_times_data_of_traces_w_spr_less_or_equal_1(
                self.mass_pos_data[self.selected_key], 'Mass Possition')
        try:
            if self.selected_key in self.soh_data.keys():
                self.combine_times_data_of_traces_w_spr_less_or_equal_1(
                    self.soh_data[self.selected_key], 'SOH')
        except KeyError:
            # Reftek's SOH trace doesn't have startTmEpoch and
            # actually soh_data consists of only one trace
            pass

    def populate_cur_key_for_all_data(self, cur_key: Tuple[str, str]) -> None:
        """
        Set up new data set's key for all data

        :param cur_key: current processing key: DAS SN, experiment number
        """
        if cur_key not in self.log_data:
            self.log_data[cur_key] = {}
            self.soh_data[cur_key] = {}
            self.mass_pos_data[cur_key] = {}
            self.waveform_data[cur_key] = {}
            self.gaps[cur_key] = []
            self.data_time[cur_key] = [constants.HIGHEST_INT, 0]
            self.stream_header_by_key_chan[cur_key] = {}

    def reset_all_selected_data(self):
        """
        Remove all keys created in the plotting process, and change fullData
        to False. This function is to replace deepcopy which uses more memory.
        """
        self.reset_data(self.soh_data[self.selected_key])
        self.reset_data(self.waveform_data[self.selected_key])
        self.reset_data(self.mass_pos_data[self.selected_key])

    def reset_data(self, data_dict: Dict):
        """
        Remove all keys created in the plotting process for the given data dict
        :param data_dict: data of the selected key
        """
        for chan_id in data_dict:
            data_dict[chan_id]['fullData'] = False
            data_dict[chan_id]['needConvert'] = True
            del_keys = ['chan_db_info', 'times', 'data', 'ax', 'ax_wf']
            for k in del_keys:
                try:
                    del data_dict[chan_id][k]
                except KeyError:
                    pass

    def reset_need_process_for_mass_pos(self):
        """
        Mass position channels are plotted twice,
            once in SOH GUI, once in waveform GUI.
        Process of trimming and down-sampling data will be conducted if
            needProcess is up.
        needProcess at the beginning is up.
        When doing the first process, put needProcess flag down
            so it won't be processed twice.
        """
        for chan_id in self.mass_pos_data[self.selected_key]:
            self.mass_pos_data[self.selected_key][chan_id][
                'needProcess'] = True
